/**************************
 * SimpleFirmModel
 * Main class for the model runs from
 * By cas220
 **************************/

/* ABM Initialization By CAS220 */
package models.SimpleFirmModel;

import models.SimpleFirmModel.parameters.Globals;
import models.SimpleFirmModel.parameters.Ranking;
import models.SimpleFirmModel.parameters.Specialization;
import models.client_company.DefaultClientCompany;
import models.client_contract.DefaultContractGenerationStrategy;
import models.client_contract.DefaultContractVisualization;
import models.consultant.JrConsultant;
import models.consultant.SrConsultant;
import models.home_company.MainFirm;
import models.market_enviroment.ConstantBusinessCycle;
import models.market_enviroment.DefaultBusinessCycle;
import models.market_enviroment.DefaultMarket;
import simudyne.core.abm.AgentBasedModel;
import simudyne.core.abm.Group;
import simudyne.core.abm.Split;
import simudyne.core.annotations.ModelSettings;

import java.util.Random;

// This Function represents the amount of ticks to progress:
@ModelSettings(macroStep = 100)
public class SimpleFirmModel extends AgentBasedModel<Globals> {

  // ABM Initializer:
  @Override
  public void init() {

    // Accumulators: (MonteCarlo Simulations)
    createLongAccumulator("DelayedContracts");
    createLongAccumulator("MonthlyRevenue");
    createLongAccumulator("MonthlySalary");
    createLongAccumulator("MonthlyGrossProfit");
    createLongAccumulator("MonthlyEBIT");
    createLongAccumulator("MonthlyNetProfit");
    createLongAccumulator("MonthlyContractProfit");

    // Register Agents:
    registerAgentTypes(
        MainFirm.class,
        DefaultClientCompany.class,
        JrConsultant.class,
        SrConsultant.class,
        DefaultContractVisualization.class,
        DefaultMarket.class);

    // Register Links (Messages):
    registerLinkTypes(
        Links.DeloitteClientLink.class,
        Links.DeloitteConsultantLink.class,
        Links.ConsultantLink.class,
        Links.ContractToClient.class,
        Links.ClientCompanyMarketLink.class,
        Links.DeloitteMarketLink.class);

    // Debugging Variable:
    createDoubleAccumulator("srEmploymentRate");
    createDoubleAccumulator("jrEmploymentRate");
  }

  // Constructor for Network and Agents:
  @Override
  public void setup() {

    // Deloitte Firm
    Group<MainFirm> mainFirmGroup =
        generateGroup(
            MainFirm.class,
            getGlobals().mainFirmAgents,
            a -> {
              a.name = "Main Firm";
            });

    // Initializing the market
    Group<DefaultMarket> marketGroup =
        generateGroup(
            DefaultMarket.class,
            1,
            a -> {
              a.name = "Market";
              a.agentID = a.getID();

              // Setting Up environment market:
              if (getGlobals().setMarketConstantGrowth) {
                a.businessCycle = new ConstantBusinessCycle(getGlobals().marketStart);
              } else {
                a.businessCycle = new DefaultBusinessCycle(getGlobals().marketStart);
              }

              // Setting Up Employment Rate:
              a.srEmploymentMean = getGlobals().srEmploymentMean;
              a.jrEmploymentMean = getGlobals().jrEmploymentMean;
            });

    // SrConsultant
    Group<SrConsultant> srConsultantGroup =
        generateGroup(
            SrConsultant.class,
            getGlobals().nbSrConsultants,
            a -> {
              a.ranking = Ranking.SENIOR;
              a.generateAllowedOverlappedProjects();
              a.generateSalary();

              // Todo: Ability to select min of agents in each discipline
              a.specialization = Specialization.generateNewRandomSpecialization();
              // a.specialization = a.assignAgentSpecialization();

              // Debugging
              a.dbAgentSpecialization = a.specialization.toString();
              a.dbAgentStatus = a.ranking.toString();
            });

    // JrConsultant
    Group<JrConsultant> jrConsultantGroup =
        generateGroup(
            JrConsultant.class,
            getGlobals().nbJrConsultants,
            a -> {
              a.ranking = Ranking.JUNIOR;
              a.generateAllowedOverlappedProjects();
              a.generateSalary();

              // Todo: Ability to select min of agents in each discipline
              a.specialization = Specialization.generateNewRandomSpecialization();
              // a.specialization = a.assignAgentSpecialization();

              // Debugging
              a.dbAgentSpecialization = a.specialization.toString();
              a.dbAgentStatus = a.ranking.toString();
            });

    // Companies List
    Group<DefaultClientCompany> clientCompanyGroup =
        generateGroup(
            DefaultClientCompany.class,
            getGlobals().nbCompanies,
            a -> {

              // Company Characteristics Setup;
              a.name = "Company # " + a.getID();

              // Company Specialization Setup:
              if (getGlobals().counter < 3) {
                a.compSpecialization = Specialization.assignNumeric(getGlobals().counter);
              } else {
                a.compSpecialization = Specialization.generateNewRandomSpecialization();
              }
              // Contract Characteristics Setup
              a.nbSimultaneousContracts = getGlobals().nbContracts + new Random().nextInt(5);
              a.contractGenerationStrategy = new DefaultContractGenerationStrategy();

              // Debugging: (Note only showing the first generated Contract)
              a.dbCompSpecialization = a.compSpecialization.toString();
              getGlobals().incrementCounter();
            });

    // Deloitte - ClientCompany links
    mainFirmGroup.fullyConnected(clientCompanyGroup, Links.DeloitteClientLink.class);
    clientCompanyGroup.fullyConnected(mainFirmGroup, Links.DeloitteClientLink.class);

    // Deloitte - SrConsultant links
    mainFirmGroup.fullyConnected(srConsultantGroup, Links.DeloitteConsultantLink.class);
    srConsultantGroup.fullyConnected(mainFirmGroup, Links.DeloitteConsultantLink.class);

    // Deloitte - JrConsultant links
    mainFirmGroup.fullyConnected(jrConsultantGroup, Links.DeloitteConsultantLink.class);
    jrConsultantGroup.fullyConnected(mainFirmGroup, Links.DeloitteConsultantLink.class);

    // Consultants - Consultants links (Can be uncommented if the computing power is enough)
    /*
       srConsultantGroup.fullyConnected(srConsultantGroup, Links.ConsultantLink.class);
       jrConsultantGroup.fullyConnected(jrConsultantGroup, Links.ConsultantLink.class);

       // SrConsultants - JrConsultants links
       jrConsultantGroup.fullyConnected(srConsultantGroup, Links.ConsultantLink.class);
       srConsultantGroup.fullyConnected(jrConsultantGroup, Links.ConsultantLink.class);
    */

    // Market - Deloitte links
    mainFirmGroup.fullyConnected(marketGroup, Links.DeloitteMarketLink.class);
    marketGroup.fullyConnected(mainFirmGroup, Links.DeloitteMarketLink.class);

    // Market - ClientCompany
    clientCompanyGroup.fullyConnected(marketGroup, Links.ClientCompanyMarketLink.class);
    marketGroup.fullyConnected(clientCompanyGroup, Links.ClientCompanyMarketLink.class);

    super.setup();
  }

  // Step function: This is a loop repeated for each step of the model.
  @Override
  public void step() {
    super.step();

    // Setup Step
    if (getContext().getTick() == 0) {
      run(
          Split.create(MainFirm.registerWithMarket, DefaultClientCompany.registerWithMarket),
          DefaultMarket.registerCompanies);
      run(
          Split.create(SrConsultant.registerWithFirm, JrConsultant.registerWithFirm),
          MainFirm.registerConsultants);
    }

    // Contract Acquisition:
    run(
        DefaultClientCompany.contractProposal,
        MainFirm.contractReview,
        Split.create(
            DefaultClientCompany.contractProposalResponse,
            SrConsultant.consultantRequest,
            JrConsultant.consultantRequest));

    // Step Though Contracts:
    run(MainFirm.contractStep);
    run(
        DefaultContractVisualization
            .stepContract); // Commented out can be used to debug code easier.

    // Terminate Contracts:
    run(
        MainFirm.terminateContracts,
        Split.create(
            DefaultClientCompany.contractCompleted,
            SrConsultant.consultantReleased,
            JrConsultant.consultantReleased));

    // Consultants Quitting:
    run(Split.create(SrConsultant.consultantQuit, JrConsultant.consultantQuit));

    // Hiring More consultants
    run(MainFirm.hireConsultants);
    if (getGlobals().hasHiredConsultants) {
      run(
          Split.create(SrConsultant.registerWithFirm, JrConsultant.registerWithFirm),
          MainFirm.registerConsultants);
      getGlobals().hasHiredConsultants = false;
    }

    // Register Monthly Revenue:
    run(
        Split.create(SrConsultant.revenueNsalarySend, JrConsultant.revenueNsalarySend),
        MainFirm.profitNloss);

    // Updating the World Market:
    run(DefaultMarket.updateEmploymentRate);
    run(DefaultMarket.updateMarketCycle);
    run(
        DefaultMarket.updateClientCompanyMarket,
        Split.create(DefaultClientCompany.registerWithMarket, DefaultClientCompany.leveHomeCompany),
        DefaultMarket.registerCompanies);

    // Iteration Global Updates:
    // Updating market rate for consultants:
    getDoubleAccumulator("srEmploymentRate").add(getGlobals().srEmploymentMean);
    getDoubleAccumulator("jrEmploymentRate").add(getGlobals().jrEmploymentMean);
  }
}
